#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='molcas_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='molcas.nomadmetainfo.json'))


class x_molcas_section_frequency(MSection):
    '''
    Section for Molcas frequency (symmetry, frequency, intensity)
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_molcas_section_frequency'))

    x_molcas_frequency_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Molcas frequency value
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_frequency_value'))

    x_molcas_imaginary_frequency_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Molcas imaginary frequency value
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_imaginary_frequency_value'))

    x_molcas_frequency_intensity = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Molcas intensity value
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_frequency_intensity'))

    x_molcas_frequency_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        Molcas symmetry for frequencies
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_frequency_symmetry'))


class x_molcas_section_basis(MSection):
    '''
    Section for Molcas basis sets
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_molcas_section_basis'))

    x_molcas_basis_atom_label = Quantity(
        type=str,
        shape=[],
        description='''
        Molcas basis set atom label.
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_basis_atom_label'))

    x_molcas_basis_name = Quantity(
        type=str,
        shape=[],
        description='''
        Molcas basis set name.  Repeated strings of '.' are compressed to a single '.'.
        Any leading or trailing '.' are stripped.
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_basis_name'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_molcas_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        Molcas method name (without UHF; see x_molcas_uhf)
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_method_name'))

    x_molcas_uhf = Quantity(
        type=bool,
        shape=[],
        description='''
        If the Molcas method is UHF.
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_uhf'))

    x_molcas_section_basis = SubSection(
        sub_section=SectionProxy('x_molcas_section_basis'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_molcas_section_basis'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_molcas_slapaf_grad_norm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Molcas slapaf (geometry optimization) grad (force) norm
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_slapaf_grad_norm'))

    x_molcas_slapaf_grad_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Molcas slapaf (geometry optimization) grad (force) max
        ''',
        a_legacy=LegacyDefinition(name='x_molcas_slapaf_grad_max'))

    x_molcas_section_frequency = SubSection(
        sub_section=SectionProxy('x_molcas_section_frequency'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_molcas_section_frequency'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os.path
import json
import ase
import numpy as np
from datetime import datetime
import time
import ast
import logging
import re

from nomad.datamodel import Author
from nomad.datamodel.metainfo.common_experimental import (
    Experiment, Sample, Method, Data, Material)

from nomad.parsing.parser import FairdiParser


logger = logging.getLogger(__name__)


class EELSApiJsonConverter(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/eels', code_name='eels', code_homepage='https://eelsdb.eu/',
            domain='ems',
            mainfile_mime_re=r'application/json',
            mainfile_contents_re=(r'https://eelsdb.eu/spectra')
        )

    def parse(self, filepath, archive, logger=logger):
        with open(filepath) as f:
            data = json.load(f)

        experiment = archive.m_create(Experiment)
        experiment.raw_metadata = data

        experiment.experiment_publish_time = datetime.strptime(
            data.get('published'), '%Y-%m-%d %H:%M:%S')

        sample = experiment.m_create(Sample)
        material = sample.m_create(Material)
        material.chemical_formula = data.get('formula')
        elements = data.get('elements')
        if elements is not None:
            if isinstance(elements, str):
                elements = json.loads(elements)
            material.atom_labels = elements
        material.chemical_name = data.get('title')

        experiment.m_create(
            Method,
            data_type='spectrum',
            method_name='electron energy loss spectroscopy',
            method_abbreviation='EELS')

        section_data = experiment.m_create(Data)

        archive.section_metadata.external_id = str(data.get('id'))
        archive.section_metadata.external_db = 'EELSDB'

        author = data.get('author')['name']
        author = re.sub(r'\(.*\)', '', author).strip()
        archive.section_metadata.coauthors = [Author(
            first_name=' '.join(author.split(' ')[0:-1]),
            last_name=author.split(' ')[-1])]
        archive.section_metadata.comment = data.get('description')
        archive.section_metadata.references = [
            data.get('permalink'),
            data.get('api_permalink'),
            data.get('download_link')
        ]

        reference = data.get('reference')
        if reference:
            if isinstance(reference, str):
                archive.section_metadata.references.append(reference)
            elif isinstance(reference, dict):
                if 'freetext' in reference:
                    archive.section_metadata.references.append(
                        re.sub(r'\r+\n+', '; ', reference['freetext']))
                else:
                    archive.section_metadata.references.append('; '.join([
                        str(value).strip() for value in reference.values()
                    ]))

        section_data.repository_name = 'Electron Energy Loss Spectroscopy (EELS) database'
        section_data.repository_url = 'https://eelsdb.eu'
        section_data.entry_repository_url = data.get('permalink')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np

from nomad.metainfo import Package, Quantity, Section
from nomad.datamodel.metainfo import common_experimental


m_package = Package(name='mpes')


class Method(common_experimental.Method):

    m_def = Section(validate=False, extends_base_section=True)

    general_beamline = Quantity(
        type=str,
        description='''
        Name of the beamline the experiment took place.
        ''')

    general_source_pump = Quantity(
        type=str,
        description='''
        Name or model of the pump light source.
        ''')

    general_source_probe = Quantity(
        type=str,
        description='''
        Name or model of the probe light source.
        ''')

    number_of_axes = Quantity(
        type=int,
        description='''
        Number of axes in the measurement hardware.
        ''')

    general_measurement_axis = Quantity(
        type=str,
        shape=['number_of_axes'],
        description='''
        Names of the axes in the measurement hardware.
        ''')

    general_physical_axis = Quantity(
        type=str,
        shape=['number_of_axes'],
        description='''
        Names of the axes in physical terms.
        ''')

    source_pump_repetition_rate = Quantity(
        type=np.dtype(np.float64),
        unit='hertz',
        description='''
        Repetition rate of the pump source.
        ''')

    source_pump_pulse_duration = Quantity(
        type=np.dtype(np.float64),
        unit='femtosecond',
        description='''
        Pulse duration of the pump source.
        ''')

    source_pump_wavelength = Quantity(
        type=np.dtype(np.float64),
        unit='nanometer',
        description='''
        Center wavelength of the pump source.
        ''')

    source_pump_spectrum = Quantity(
        type=np.dtype(np.float64),
        shape=['length_of_spectrum'],
        description='''
        Spectrum of the pump source.
        ''')

    source_pump_photon_energy = Quantity(
        type=np.dtype(np.float64),
        unit='electron_volt',
        description='''
        Photon energy of the pump source.
        ''')

    source_pump_size = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='millimeter ** 2',
        description='''
        Full-width at half-maximum (FWHM) of the pump source size at or closest to the
        sample position.
        ''')

    source_pump_fluence = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='millijoule / millimeter ** 2',
        description='''
        Fluence of the pump source at or closest to the sample position.
        ''')

    source_pump_polarization = Quantity(
        type=str,
        description='''
        Polarization of the pump source.
        ''')

    source_pump_bunch = Quantity(
        type=np.dtype(np.int32),
        description='''
        Total bunch number of the pump source.
        ''')

    source_probe_repetition_rate = Quantity(
        type=np.dtype(np.float64),
        unit='hertz',
        description='''
        Repetition rate of the probe source.
        ''')

    source_probe_pulse_duration = Quantity(
        type=np.dtype(np.float64),
        unit='femtosecond',
        description='''
        Pulse duration of the probe source.
        ''')

    source_probe_wavelength = Quantity(
        type=np.dtype(np.float64),
        unit='nanometer',
        description='''
        Center wavelength of the probe source.
        ''')

    length_of_spectrum = Quantity(
        type=int,
        description='''
        Number of pixel elements in the spectrum.
        ''')

    source_probe_spectrum = Quantity(
        type=np.dtype(np.float64),
        shape=['length_of_spectrum'],
        description='''
        Spectrum of the probe source.
        ''')

    source_probe_photon_energy = Quantity(
        type=np.dtype(np.float64),
        unit='electron_volt',
        description='''
        Photon energy of the probe source.
        ''')

    source_probe_size = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='millimeter ** 2',
        description='''
        Full-width at half-maximum (FWHM) of the probe source size at or closest to the
        sample position.
        ''')

    source_probe_fluence = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='millijoule / millimeter ** 2',
        description='''
        Fluence of the probe source at or closest to the sample position.
        ''')

    source_probe_polarization = Quantity(
        type=str,
        description='''
        Polarization of the probe source.
        ''')

    source_probe_bunch = Quantity(
        type=np.dtype(np.int32),
        description='''
        Total bunch number of the probe source.
        ''')

    source_temporal_resolution = Quantity(
        type=np.dtype(np.float64),
        unit='femtosecond',
        description='''
        Full-width at half-maximum (FWHM) of the pump-probe cross-correlation function.
        ''')

    detector_extractor_voltage = Quantity(
        type=np.dtype(np.float64),
        unit='volt',
        description='''
        Voltage between the extractor and the sample.
        ''')

    detector_work_distance = Quantity(
        type=np.dtype(np.float64),
        unit='millimeter',
        description='''
        Distance between the sample and the detector entrance.
        ''')

    number_of_lenses = Quantity(
        type=int,
        description='''
        Number of electron lenses in the electron detector.
        ''')

    detector_lens_names = Quantity(
        type=str,
        shape=['number_of_lenses'],
        description='''
        Set of names for the electron-optic lenses.
        ''')

    detector_lens_voltages = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_lenses'],
        unit='volt',
        description='''
        Set of electron-optic lens voltages.
        ''')

    detector_tof_distance = Quantity(
        type=np.dtype(np.float64),
        unit='meter',
        description='''
        Drift distance of the time-of-flight tube.
        ''')

    number_of_tof_voltages = Quantity(
        type=int,
        description='''
        Number of time-of-flight (TOF) drift tube voltage values in the electron detector.
        ''')

    detector_tof_voltages = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_tof_voltages'],
        unit='volt',
        description='''
        Voltage applied to the time-of-flight tube.
        ''')

    detector_sample_bias = Quantity(
        type=np.dtype(np.float64),
        unit='volt',
        description='''
        Voltage bias applied to sample.
        ''')

    detector_magnification = Quantity(
        type=np.dtype(np.float64),
        description='''
        Detector magnification.
        ''')

    number_of_detector_voltages = Quantity(
        type=int,
        description='''
        Number of detector voltage settings in the electron detector.
        ''')

    detector_voltages = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_detector_voltages'],
        unit='volt',
        description='''
        Voltage applied to detector.
        ''')

    detector_type = Quantity(
        type=str,
        description='''
        Description of the detector type (e.g. ‘MCP’, ‘CCD’, ‘CMOS’, etc.).
        ''')

    number_of_sensor_sizes = Quantity(
        type=int,
        description='''
        Number of detector sensor size dimensions (depending on the number of sensors).
        ''')

    detector_sensor_size = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_sensor_sizes'],
        unit='millimeter',
        description='''
        Size of each of the imaging sensor chip on the detector.
        ''')

    detector_sensor_count = Quantity(
        type=np.dtype(np.int32),
        description='''
        Number of imaging sensor chips on the detector.
        ''')

    detector_sensor_pixel_size = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='micrometer',
        description='''
        Pixel size of the imaging sensor chip on the detector.
        ''')

    number_of_momentum_calibration_coefficients = Quantity(
        type=int,
        description='''
        Number of the momentum calibration parameters for the detector.
        ''')

    detector_calibration_x_to_momentum = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_momentum_calibration_coefficients'],
        unit='1 / angstrom',
        description='''
        Pixel x axis to kx momentum calibration.
        ''')

    detector_calibration_y_to_momentum = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_momentum_calibration_coefficients'],
        unit='1 / angstrom',
        description='''
        Pixel y axis to ky momentum calibration.
        ''')

    number_of_energy_calibration_coefficients = Quantity(
        type=int,
        description='''
        Number of the energy calibration parameters for the detector.
        ''')

    detector_calibration_tof_to_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_energy_calibration_coefficients'],
        unit='electron_volt',
        description='''
        Time-of-flight to energy calibration.
        ''')

    detector_calibration_stage_to_delay = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_delay_calibration_coefficients'],
        unit='femtosecond',
        description='''
        Translation stage position to pump-probe delay calibration.
        ''')

    number_of_other_calibration_coefficients = Quantity(
        type=int,
        description='''
        Number of the other calibration parameters for the detector.
        ''')

    detector_calibration_other_converts = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_other_calibration_coefficients'],
        description='''
        Conversion factor between other measured and physical axes.
        ''')

    detector_momentum_resolution = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='1 / angstrom',
        description='''
        Momentum resolution of the detector.
        ''')

    detector_spatial_resolution = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='micrometer',
        description='''
        Spatial resolution of the source.
        ''')

    detector_energy_resolution = Quantity(
        type=np.dtype(np.float64),
        shape=['none_shape'],
        unit='electron_volt',
        description='''
        Energy resolution of the detector.
        ''')


class Sample(common_experimental.Sample):

    sample_state_of_matter = Quantity(
        type=str,
        description='''
        Physical state of the sample.
        ''')

    sample_purity = Quantity(
        type=np.dtype(np.float64),
        description='''
        Chemical purity of the sample.
        ''')

    sample_surface_termination = Quantity(
        type=str,
        description='''
        Surface termination of the sample (if crystalline).
        ''')

    sample_layers = Quantity(
        type=str,
        description='''
        Sample layer or bulk structure.
        ''')

    sample_stacking_order = Quantity(
        type=str,
        description='''
        Stacking order of the solid surface (if crystalline).
        ''')

    sample_chemical_id_cas = Quantity(
        type=str,
        description='''
        CAS registry number of the sample’s chemical content.
        ''')

    sample_pressure = Quantity(
        type=np.dtype(np.float64),
        unit='pascal',
        description='''
        Pressure surrounding the sample at the time of measurement.
        ''')

    sample_growth_method = Quantity(
        type=str,
        description='''
        Sample growth method.
        ''')

    sample_preparation_method = Quantity(
        type=str,
        description='''
        Sample preparation method.
        ''')

    sample_vendor = Quantity(
        type=str,
        description='''
        Name of the sample vendor.
        ''')

    sample_substrate_material = Quantity(
        type=str,
        description='''
        Material of the substrate the sample has immediate contact with.
        ''')

    sample_substrate_state_of_matter = Quantity(
        type=str,
        description='''
        State of matter of the substrate material.
        ''')

    sample_substrate_vendor = Quantity(
        type=str,
        description='''
        Name of the substrate vendor.
        ''')


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click

from nomad import processing as proc, infrastructure, search
from .admin import admin


@admin.group(help='Entry related commands')
def entries():
    infrastructure.setup_mongo()
    infrastructure.setup_elastic()


@entries.command(help='Delete selected entries from mongo and elastic')
@click.argument('ENTRIES', nargs=-1)
@click.option('--skip-es', help='Keep the elastic index version of the data.', is_flag=True)
@click.option('--skip-mongo', help='Keep uploads and calcs in mongo.', is_flag=True)
def rm(entries, skip_es, skip_mongo):
    print('%d entries selected, deleting ...' % len(entries))

    if not skip_es:
        for entry in entries:
            search.delete_entry(calc_id=entry)

    if not skip_mongo:
        proc.Calc.objects(calc_id__in=entries).delete()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Processing comprises everything that is necessary to take an uploaded user file,
processes it, and store all necessary data for *repository*, *archive*, and potential
future services (e.g. *encyclopedia*).

Processing is build on top of *celery* (http://www.celeryproject.org/) and
*mongodb* (http://www.mongodb.org).
Celery provides a task-based programming model for distributed computing. It uses
a broker, e.g. a distributed task queue like *RabbitMQ* to distribute tasks. We
use mongodb to store the current state of processing in :class:`Upload` and
:class:`Calculation` documents. This combination allows us to easily distribute
processing work while having the processing state, i.e. (intermediate) results,
always available.

This module is structured into our *celery app* and abstract process base class
:class:`Proc` (``base.py``), and the concrete processing classes
:class:`Upload` and :class:`Calc` (``data.py``).

This module does not contain the functions to do the actual work. Those are encapsulated
in :py:mod:`nomad.files`, :py:mod:`nomad.repo`, :py:mod:`nomad.users`,
:py:mod:`nomad.parsing`, and :py:mod:`nomad.normalizing`.

Refer to http://www.celeryproject.org/ to learn about celery apps and workers. The
nomad celery app uses a *RabbitMQ* broker. We use celery to distribute processing load
in a cluster.

We use an abstract processing base class and document (:class:`Proc`) that provides all
necessary functions to execute a process as a series of potentially distributed steps. In
addition the processing state is persisted in mongodb using *mongoengine*. Instead of
exchanging serialized state between celery tasks, we use the mongodb documents to
exchange data. Therefore, the mongodb always contains the latest processing state.
We also don't have to deal with celery result backends and synchronizing with them.

.. autoclass:: nomad.processing.base.Proc

There are two concrete processes :class:`Upload` and :class: `Calc`. Instances of both
classes do represent the processing state, as well as the respective entity.

.. autoclass:: nomad.processing.data.Upload
    :members:
.. autoclass:: nomad.processing.data.Calc
    :members:
'''

from nomad.processing.base import app, InvalidId, ProcNotRegistered, SUCCESS, FAILURE, \
    RUNNING, PENDING, PROCESS_COMPLETED, PROCESS_RUNNING, ProcessAlreadyRunning
from nomad.processing.data import Upload, Calc



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import mmap
import io
import re
import numpy as np
import pint
from typing import List, Union, Callable, Type

from nomad.parsing.file_parser import FileParser
from nomad.metainfo import Quantity as mQuantity


class ParsePattern:
    def __init__(self, **kwargs):
        self._head = kwargs.get('head', '')
        self._key = kwargs.get('key', '')
        value = kwargs.get('value', 're_float_array')
        if value.startswith('re_'):
            token = ''
            if 'float' in value:
                token += r'Ee\+\d\.\-'
            if 'int' in value:
                token += r'\d'
            if 'str' in value:
                token += r'\w'
            if 'array' in value:
                token += r' '
            value = r'[%s]+' % token
        self._value = value
        self._tail = kwargs.get('tail', '\n')
        self._re_pattern = None

    @property
    def re_pattern(self):
        if self._re_pattern is None:
            head = r'%s[\s\S]*?' % self._head if self._head else ''
            key = r'%s\s*\:*\=*\s*' % self._key if self._key else ''
            self._re_pattern = r'%s%s\s*\:*\=*\s*(%s)%s' % (
                head, key, self._value, self._tail)
        return self._re_pattern

    def __call__(self, text, repeats=True):
        values = []
        units = []
        if repeats:
            for res in self.re_pattern.finditer(text):
                unit = res.groupdict().get('__unit', None)
                values.append(
                    ''.join([group.decode() for group in res.groups() if group and group != unit]))
                units.append(unit.decode() if unit is not None else None)
        else:
            res = self.re_pattern.search(text)
            if res is not None:
                unit = res.groupdict().get('__unit', None)
                units.append(unit.decode() if unit is not None else None)
                values.append(''.join(
                    [group.decode() for group in res.groups() if group and group != unit]))


class Quantity:
    '''
    Class to define a quantity to be parsed in the TextParser.

    Arguments:
        quantity: string to identify the name or a metainfo quantity to initialize the
            quantity object.
        re_pattern: pattern to be used by re for matching. Ideally, overlaps among
            quantities for a given parser should be avoided.
        sub_parser: instance of TextParser to perform local parsing
            within a matched block
        str_operation: external function to be performed on a matched block
        dtype: data type of the quantity
        unit: unit of the quantity
        shape: shape of the quantity
        repeats: denotes if multiple matches are expected
        convert: switch automatic data type conversion
        comment: character to denote a line to be ignored

    '''
    def __init__(self, quantity: Union[str, mQuantity], re_pattern: Union[str, ParsePattern], **kwargs):
        self.name: str
        self.dtype: str
        self.unit: str
        self.shape: List[int]
        if isinstance(quantity, str):
            self.name = quantity
            self.dtype = None
            self.unit = None
            self.shape = None
        elif isinstance(quantity, mQuantity):
            self.name = quantity.name
            self.dtype = quantity.type.type if isinstance(quantity.type, np.dtype) else quantity.type
            self.unit = quantity.unit
            # check if metainfo shape has dependencies
            self.shape = quantity.shape
            if False in [str(i).isdigit() for i in self.shape]:
                self.shape = None
        # override metainfo
        self.dtype = kwargs.get('dtype', self.dtype)
        self.unit = kwargs.get('unit', self.unit)
        self.shape = kwargs.get('shape', self.shape)

        self._re_pattern: str = re_pattern.re_pattern if isinstance(
            re_pattern, ParsePattern) else re_pattern
        self._str_operation: Callable = kwargs.get('str_operation', None)
        self._sub_parser: TextParser = kwargs.get('sub_parser', None)
        self.repeats: bool = kwargs.get('repeats', False)
        self.convert: bool = kwargs.get('convert', True)
        self.flatten: bool = kwargs.get('flatten', True)
        self.comment: str = kwargs.get('comment', None)

    @property
    def re_pattern(self):
        '''
        Returns a compiled re pattern.
        '''
        if isinstance(self._re_pattern, str):
            re_pattern = self._re_pattern.replace('__unit', '__unit_%s' % self.name)
            self._re_pattern = re.compile(re_pattern.encode())
        return self._re_pattern

    @re_pattern.setter
    def re_pattern(self, val: str):
        self._re_pattern = val

    @property
    def str_operation(self):
        return self._str_operation

    @str_operation.setter
    def str_operation(self, val: Callable):
        self._str_operation = val

    def to_data(self, val_in: List[str]):
        '''
        Converts the parsed block into data.
        '''
        def process(val):
            if self.comment is not None:
                if val.strip()[0] == self.comment:
                    return

            if self.str_operation is not None:
                val = self.str_operation(val)

            elif self.flatten:
                val = val.strip().split() if isinstance(val, str) else val
                val = val[0] if len(val) == 1 else val

            def _convert(val):
                if isinstance(val, str):
                    if self.dtype is None:
                        if val.isdecimal():
                            val = int(val)
                        else:
                            try:
                                val = float(val)
                            except Exception:
                                pass
                    else:
                        try:
                            val = self.dtype(val)
                        except Exception:
                            pass

                    return val

                elif type(val) in [np.ndarray, list]:
                    try:
                        dtype = float if self.dtype is None else self.dtype
                        val_test = np.array(val, dtype=dtype)
                        if self.dtype is None:
                            if np.all(np.mod(val_test, 1) == 0):
                                val_test = np.array(val_test, dtype=int)
                        val = val_test

                    except Exception:
                        val = [_convert(v) for v in val]

                    return val

                elif isinstance(val, dict):
                    for k, v in val.items():
                        val[k] = _convert(v)
                    return val

                else:
                    return val

            if self.convert:
                val = _convert(val)

            if isinstance(val, np.ndarray) and self.shape:
                try:
                    val = np.reshape(val, self.shape)
                except Exception:
                    pass

            return val

        val_out = [process(val) for val in val_in]

        if isinstance(val_out[0], np.ndarray):
            self.dtype = val_out[0].dtype  # type: ignore

        return val_out


class DataTextParser(FileParser):
    '''
    Parser for structured data text files using numpy.loadtxt

    Arguments:
        mainfile: the file to be parsed
        dtype: data type
    '''
    def __init__(self, **kwargs):
        self._dtype: Type = kwargs.get('dtype', float)
        mainfile: str = kwargs.get('mainfile', None)
        self._mainfile_contents: str = kwargs.get('mainfile_contents', None)
        logger = kwargs.get('logger', None)
        logger = logger if logger is not None else logging
        super().__init__(mainfile, logger=logger, open=kwargs.get('open', None))
        self.init_parameters()

    def init_parameters(self):
        '''
        Method to call after loading data.
        '''
        pass

    @property
    def data(self):
        '''
        Returns the loaded data
        '''
        if self._file_handler is None:
            try:
                if self.mainfile is not None:
                    self._file_handler = np.loadtxt(self.mainfile)
                else:
                    if self._mainfile_contents is None:
                        self._mainfile_contents = self.mainfile_obj.read()
                    self._file_handler = np.frombuffer(self._mainfile_contents, dtype=self._dtype)
            except Exception:
                return

            self.init_parameters()
        return self._file_handler


class TextParser(FileParser):
    '''
    Parser for unstructured text files using the re module. The quantities to be parsed
    are given as a list of Quantity objects which specifies the re pattern. The mmap
    module is used to handle the file. By default, re.find_all is used to get matches
    for performance reasons. In this case, overlap is not tolerated in the re patterns.
    To avoid this, set findall to False to switch to re.finditer.

    Arguments:
        mainfile: the file to be parsed
        quantities: list of Quantity objects to be parsed.
        logger: optional logger
        findall: switches between using re.findall and re.finditer
        file_offset: offset in reading the file
        file_length: length of the chunk to be read from the file
    '''
    def __init__(self, mainfile=None, quantities=None, logger=None, findall=True, **kwargs):
        super().__init__(mainfile, logger=logger, open=kwargs.get('open', None))
        self._quantities: List[Quantity] = quantities
        self.findall: bool = findall
        self._kwargs = kwargs
        self._file_length: int = kwargs.get('file_length', 0)
        self._file_offset: int = kwargs.get('file_offset', 0)
        self._file_pad: int = 0
        if quantities is None:
            self.init_quantities()
        # check quantity patterns are valid
        re_has_group = re.compile(r'\(.+\)')
        for i in range(len(self.quantities) - 1, -1, -1):
            valid = True
            try:
                if re_has_group.search(self.quantities[i].re_pattern.pattern.decode()) is None:
                    valid = False
            except Exception:
                valid = False
            if not valid:
                self.logger.error(
                    'Invalid quantity pattern', data=dict(quantity=self.quantities[i].name))
                self.quantities.pop(i)
        self._re_findall = None

    def copy(self):
        '''
        Returns a copy of the object excluding the parsed results.
        '''
        return TextParser(
            self.mainfile, self.quantities, self.logger, **self._kwargs)

    def init_quantities(self):
        '''
        Initializes the quantities list.
        '''
        self._quantities = []

    @property
    def quantities(self):
        return self._quantities

    @quantities.setter
    def quantities(self, val):
        self._quantities = val

    @property
    def file_offset(self):
        '''
        Integer offset in loading the file taking into account mmap pagination.
        '''
        return self._file_offset

    @file_offset.setter
    def file_offset(self, val):
        self._file_pad = val % mmap.PAGESIZE
        self._file_offset = (val // mmap.PAGESIZE) * mmap.PAGESIZE

    @property
    def file_length(self):
        '''
        Length of the file chunk to be loaded.
        '''
        return self._file_length

    @file_length.setter
    def file_length(self, val):
        self._file_length = val

    @property
    def file_mmap(self):
        '''
        Memory mapped representation of the file.
        '''
        if self._file_handler is None:
            with self.open(self.mainfile) as f:
                if isinstance(f, io.TextIOWrapper):
                    self._file_handler = mmap.mmap(
                        f.fileno(), self._file_length, access=mmap.ACCESS_COPY,
                        offset=self._file_offset)
                    # set the extra chunk loaded before the intended offset to empty
                    self._file_handler[:self._file_pad] = b' ' * self._file_pad
                else:
                    self._file_handler = f.read()
            self._file_pad = 0
        return self._file_handler

    def keys(self):
        '''
        Returns all the quantity names.
        '''
        return [quantity.name for quantity in self.quantities]

    def items(self):
        '''
        Returns an iterable name, value of the parsed quantities
        '''
        for key in self.keys():
            yield key, self.get(key)

    def _parse_quantities(self, quantities):
        if len(self._results) == 0 and self._re_findall is not None:
            # maybe an opt
            re_findall = self._re_findall
        else:
            re_findall = '|'.join([q.re_pattern.pattern.decode() for q in quantities])
            if len(quantities) == 1:
                # necessary to add a dummy variable to make multiple matches
                re_findall = '%s|(__dummy__)' % re_findall
            re_findall = re_findall.encode()
            if self._re_findall is None:
                self._re_findall = re.compile(re_findall)

        # map matches to quantities
        matches = re.findall(re_findall, self.file_mmap)
        current_index = 0
        for i in range(len(quantities)):
            values = []
            units = []
            n_groups = quantities[i].re_pattern.groups

            non_empty_matches = []
            for match in matches:
                non_empty_match = [m for m in match[current_index: current_index + n_groups] if m]
                if not non_empty_match:
                    continue
                non_empty_matches.append(non_empty_match)
            index_unit = quantities[i].re_pattern.groupindex.get(
                '__unit_%s' % quantities[i].name, None)
            for non_empty_match in non_empty_matches:
                if index_unit is not None:
                    unit = non_empty_match.pop(index_unit - 1)
                    units.append(unit.decode())

                else:
                    units.append(None)

                values.append(' '.join([m.decode() for m in non_empty_match]))

            current_index += n_groups

            if not values:
                continue

            try:
                value_processed = quantities[i].to_data(values)
                for j in range(len(value_processed)):
                    unit = units[j] if units[j] else quantities[i].unit
                    if not unit:
                        continue
                    if isinstance(unit, str):
                        value_processed[j] = pint.Quantity(value_processed[j], unit)
                    else:
                        value_processed[j] = value_processed[j] * unit

                if not quantities[i].repeats and value_processed:
                    value_processed = value_processed[0]

                self._results[quantities[i].name] = value_processed

            except Exception:
                self.logger.warn('Error setting value', data=dict(quantity=quantities[i].name))
                pass

    def _parse_quantity(self, quantity):

        value = []
        units = []
        re_matches = quantity.re_pattern.finditer(self.file_mmap) if quantity.repeats else [
            quantity.re_pattern.search(self.file_mmap)]
        for res in re_matches:
            if res is None:
                continue
            if quantity._sub_parser is not None:
                span = np.array(res.span()) + self.file_offset
                sub_parser = quantity._sub_parser.copy()
                sub_parser.mainfile = self.mainfile
                sub_parser.logger = self.logger
                if (span[1] - span[0]) < mmap.PAGESIZE or True:
                    # self.logger.warn(
                    #     'Cannot use sub parser on quantity %s with blocks with size <'
                    #     '%d. Will try to parse string' % (quantity.name, mmap.PAGESIZE))
                    sub_parser._file_handler = b' '.join([g for g in res.groups() if g])
                else:
                    sub_parser.file_offset = span[0]
                    sub_parser.file_length = span[1] - sub_parser.file_offset
                value.append(sub_parser.parse())

            else:
                unit = res.groupdict().get('__unit_%s' % quantity.name, None)
                units.append(unit.decode() if unit is not None else None)
                value.append(' '.join(
                    [group.decode() for group in res.groups() if group and group != unit]))

        if not value:
            return

        if quantity._sub_parser is not None:
            self._results[quantity.name] = value if quantity.repeats else value[0]

        else:
            try:
                value_processed = quantity.to_data(value)
                for i in range(len(value_processed)):
                    unit = units[i] if units[i] else quantity.unit
                    if not unit:
                        continue
                    if isinstance(unit, str):
                        value_processed[i] = pint.Quantity(value_processed[i], unit)
                    else:
                        value_processed[i] = value_processed[i] * unit

                if not quantity.repeats and value_processed:
                    value_processed = value_processed[0]

                self._results[quantity.name] = value_processed
            except Exception:
                self.logger.warn('Error setting value', data=dict(quantity=quantity.name))
                pass

    def parse(self, key=None):
        '''
        Triggers parsing of all quantities if key is not provided.
        '''
        if self._results is None:
            self._results = dict()

        if self.file_mmap is None:
            return self

        if self.findall:
            if len(self._results) > 1:
                return self

            n_results = 0
            while True:
                quantities_findall = [
                    q for q in self.quantities if q.name not in self._results and q._sub_parser is None]
                if not quantities_findall:
                    break

                # recursively parse quantities
                self._parse_quantities(quantities_findall)

                if n_results == len(self._results):
                    break
                n_results = len(self._results)

            for quantity in self._quantities:
                if quantity._sub_parser is not None:
                    self._parse_quantity(quantity)

        else:
            for quantity in self._quantities:
                if quantity.name == key or key is None:
                    if quantity.name not in self._results:
                        self._parse_quantity(quantity)

        return self



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='lib_atoms_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='lib_atoms.nomadmetainfo.json'))


class x_lib_atoms_section_gap(MSection):
    '''
    Description of Gaussian Approximation Potentials (GAPs).
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_lib_atoms_section_gap'))

    x_lib_atoms_training_config_refs = Quantity(
        type=public.section_single_configuration_calculation,
        shape=['n_sparseX'],
        description='''
        References to frames in training configuration.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_training_config_refs'))

    x_lib_atoms_GAP_params_label = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_GAP_params_label'))

    x_lib_atoms_GAP_params_svn_version = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_GAP_params_svn_version'))

    x_lib_atoms_GAP_data_do_core = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_GAP_data_do_core'))

    x_lib_atoms_GAP_data_e0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_GAP_data_e0'))

    x_lib_atoms_command_line_command_line = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_command_line_command_line'))

    x_lib_atoms_gpSparse_n_coordinate = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpSparse_n_coordinate'))

    x_lib_atoms_gpCoordinates_n_permutations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_n_permutations'))

    x_lib_atoms_gpCoordinates_sparsified = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_sparsified'))

    x_lib_atoms_gpCoordinates_signal_variance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_signal_variance'))

    x_lib_atoms_gpCoordinates_label = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_label'))

    x_lib_atoms_gpCoordinates_n_sparseX = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_n_sparseX'))

    x_lib_atoms_gpCoordinates_covariance_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_covariance_type'))

    x_lib_atoms_gpCoordinates_signal_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_signal_mean'))

    x_lib_atoms_gpCoordinates_sparseX_filename = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_sparseX_filename'))

    x_lib_atoms_gpCoordinates_dimensions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_dimensions'))

    x_lib_atoms_gpCoordinates_theta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_theta'))

    x_lib_atoms_gpCoordinates_descriptor = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_descriptor'))

    x_lib_atoms_gpCoordinates_perm_permutation = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_perm_permutation'))

    x_lib_atoms_gpCoordinates_perm_i = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_perm_i'))

    x_lib_atoms_gpCoordinates_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sparseX', 2],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_alpha'))

    x_lib_atoms_gpCoordinates_sparseX = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sparseX', 'dimensions'],
        description='''
        GAP classifier.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_gpCoordinates_sparseX'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_lib_atoms_section_gap = SubSection(
        sub_section=SectionProxy('x_lib_atoms_section_gap'),
        repeats=False,
        a_legacy=LegacyDefinition(name='x_lib_atoms_section_gap'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_lib_atoms_virial_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='pascal',
        description='''
        Virial tensor for this frame.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_virial_tensor'))

    x_lib_atoms_config_type = Quantity(
        type=str,
        shape=[],
        description='''
        Configuration type, e.g. = dislocation_quadrupole.
        ''',
        a_legacy=LegacyDefinition(name='x_lib_atoms_config_type'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
from datetime import datetime
from ase.data import atomic_numbers

from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity, FileParser
from nomad.datamodel.metainfo.common_dft import Run, Method, MethodAtomKind, MethodBasisSet,\
    BasisSetAtomCentered, BasisSet, System, SingleConfigurationCalculation, XCFunctionals,\
    SystemToSystemRefs, CalculationToCalculationRefs, ScfIteration, Eigenvalues,\
    SamplingMethod, EnergyVanDerWaals
from turbomoleparser.metainfo.turbomole import x_turbomole_section_eigenvalues_GW
from turbomoleparser.metainfo import m_env


MOL = 6.02214076E23


class AuxiliaryOutParser(FileParser):
    def __init__(self, re_line, start):
        super().__init__()
        self._re_line = re.compile(re_line)
        self._start = start

    @property
    def matrix(self):
        if self._file_handler is None:
            self._file_handler = []
            with open(self.mainfile) as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    line = self._re_line.match(line)
                    if not line:
                        continue
                    value = line.group(1).split()
                    if int(value[0]) > len(self._file_handler):
                        self._file_handler.append([])
                    self._file_handler[-1].extend(value[self._start:])
        return self._file_handler


class EigenvaluesParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'irrep',
                r'irrep\s*([\w ]+\n)'
            ),
            Quantity(
                'eigenvalues',
                r'eigenvalues H\s*([\d\.\-\+EeDd ]+)\n',
                dtype=float, repeats=True,
                str_operation=lambda x: x.replace('D', 'E').replace('E', 'e').split()),
            Quantity(
                'occupation',
                r'occupation\s*([\d\.\- ]*\n)', dtype=float, repeats=True)]


class ControlParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_drvopt(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            return {v[0].strip(): v[1] for v in val}

        # TODO add more quantities
        self._quantities = [
            Quantity('operating_system', r'operating system\s*(\S+)'),
            Quantity('scf_iter_limit', r'scfiterlimit\s*(\S+)'),
            Quantity('scf_conv', r'scfconv\s*(\S+)'),
            Quantity('time_for_integral_calc', r'thime\s*(\S+)'),
            Quantity('pople_kind', r'pople\s*(\S+)'),
            Quantity('damping_parameter', r'scfdamp\s*start=\s*(\S+)\s*step=\s*(\S+)\s*min=\s*(\S+)'),
            Quantity('scfint', r'unit=\s*(\S+)\s*size=\s*(\S+)\s*file=\s*(\S+)\s*'),
            Quantity('interconversion_status', r'interconversion\s*(\S+)'),
            Quantity('drvopt', r'drvopt\s*([\s\S]+?)\n *\$', str_operation=str_to_drvopt)]


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+eEDd]+'

        def str_to_atomic_info(val_in):
            vals = [v.split() for v in val_in.strip().split('\n')]
            # first two are 'atomic', 'coordinates'
            if vals[0][0] == '1':
                vals = np.transpose(vals)
                info = dict(
                    coordinates=np.array(vals[-3:].T, dtype=float) * ureg.bohr, atom=vals[1])
            else:
                keys = vals[0][2:]
                vals = np.transpose(vals[1:])
                info = dict(coordinates=np.array(vals[:3].T, dtype=float) * ureg.bohr)
                for n, key in enumerate(keys):
                    info[key] = vals[n + 3]
            info['atom'] = [v.title() for v in info.get('atom', [])]
            return info

        def str_to_cpu_time(val_in):
            time = 0.
            for n, v in enumerate(reversed(val_in.split())):
                multiplier = 24 * 3600 if n == 3 else 60 ** n
                time += float(v) * multiplier
            return time

        def str_to_hessian(val_in):
            re_row = re.compile(r'(\d*)\s*\w*\s*d(?:x|y|z) +([\d\.\- ]+\d)')
            re_value = re.compile(r'\-*\d+\.\d+')
            n_axis = 3

            hessian = []
            row_n = -1
            for val in val_in.strip().split('\n'):
                row = re.search(re_row, val)
                if not row:
                    continue
                if row.group(1):
                    row_n = (int(row.group(1)) - 1) * n_axis
                else:
                    row_n += 1

                if len(hessian) < row_n + 1:
                    hessian.append([])
                hessian[row_n].extend(re_value.findall(row.group(2)))

            for n in range(len(hessian)):
                hessian[n].extend([0] * (len(hessian) - len(hessian[n])))

            hessian = np.array(hessian, dtype=float)
            # symmetrize
            hessian = hessian + hessian.T - np.diag(hessian.diagonal())
            n_atoms = len(hessian) // n_axis
            hessian = np.reshape(hessian, (n_atoms, n_axis, n_atoms, n_axis))
            return np.transpose(hessian, axes=(0, 2, 1, 3)) * ureg.hartree / ureg.bohr**2

        # TODO integrate with hessian
        def str_to_normal_modes_vibrational_frequencies(val_in):
            re_row = re.compile(r'(\d*)\s*\w*\s+(?:x|y|z) +([\d\.\- ]+\d)')
            re_value = re.compile(r'\-*\d+\.\d+')
            n_axis = 3

            row_n = -1
            normal_modes = []
            for val in val_in.strip().split('\n'):
                row = re.search(re_row, val)
                if not row:
                    continue
                if row.group(1):
                    row_n = (int(row.group(1)) - 1) * n_axis
                else:
                    row_n += 1

                if len(normal_modes) < row_n + 1:
                    normal_modes.append([])
                normal_modes[row_n].extend(re_value.findall(row.group(2)))

            n_atoms = len(normal_modes) // n_axis
            normal_modes = np.array(normal_modes, dtype=float)
            n_modes = np.size(normal_modes) // (n_atoms * n_axis)
            normal_modes = np.transpose(
                np.reshape(normal_modes, (n_atoms, n_axis, n_modes)), axes=(2, 0, 1))

            def get_values(pattern, dtype=str):
                values_str = re.findall(pattern, val_in)
                values = []
                for value in values_str:
                    if dtype == float:
                        values.extend(re_value.findall(value))
                    else:
                        values.extend(value.split())
                return np.array(values, dtype=dtype)

            frequencies = get_values(r'frequency\s+([\d\.\- ]+\d)', float)

            infrared = [v == 'YES' for v in get_values(r'IR\s+([YES\- ]+)')]

            raman = [v == 'YES' for v in get_values(r'RAMAN\s+([YES\- ]+)')]

            intensity = get_values(r'intensity \(km/mol\)\s+([\d\.\- ]+\d)', float)

            return dict(
                normal_modes=normal_modes, mode_energies=frequencies, infrared_activity=infrared,
                raman_activity=raman, intensities=intensity)

        def str_to_energy_gradient(val_in):
            re_row = re.compile(r'dE/d((?:x|y|z)) +([\d\.\-\+ Dd]+\d)')
            re_value = re.compile(r'\-*\d+\.\d+D*d*\-*\+*\d*')

            gradients = []
            for val in val_in.strip().split('\n'):
                row = re.search(re_row, val)
                if not row:
                    continue
                row_n = ['x', 'y', 'z'].index(row.group(1))
                if len(gradients) < row_n + 1:
                    gradients.append([])
                gradients[row_n].extend([
                    v.replace('D', 'E').replace('d', 'e') for v in re_value.findall(row.group(2))])

            gradients = np.array(gradients, dtype=float)
            return np.transpose(gradients) * (ureg.hartree / ureg.bohr)

        def str_to_iteration(val_in):
            val = val_in.split('\n')
            keys = [v.strip() for v in val[0].split('  ') if v]
            # replace key with metainfo name
            units = {'energy_total': ureg.hartree}
            for n in range(len(keys)):
                if keys[n] == 'CCSD energy':
                    keys[n] = 'energy_total'
                elif keys[n] == 'wall':
                    keys[n] = 'time'
            val = [v.split() for v in val[1:]]
            val = np.transpose(np.array([v for v in val if len(v) == len(keys)], dtype=float))
            return [{key: val[n][i] * units.get(
                key, 1.0) for n, key in enumerate(keys)} for i in range(len(val[0]))]

        def str_to_embedding_coordinates(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            val = np.transpose([v for v in val if len(v) >= 4])
            label = [v.title() for v in val[0]]
            charge = val[4] if len(val) > 4 else [0] * len(label)
            return dict(
                label=label, coordinates=np.array(val[1:4].T, dtype=float) * ureg.bohr,
                charge=np.array(charge, dtype=float) * ureg.elementary_charge)

        def str_to_scf_energies(val_in):
            val = [float(v) for v in val_in.strip().replace('D', 'E').split()]
            return dict(
                energy_total_scf_iteration=val[1] * ureg.hartree,
                x_turbomole_energy_1electron_scf_iteration=val[2] * ureg.hartree,
                x_turbomole_energy_2electron_scf_iteration=val[3] * ureg.hartree)

        def str_to_norm(val_in):
            name, val = val_in.split('=')
            val = val.split()
            name = 'norm_fia' if 'Fia' in name else 'norm_fock'
            return {name: float(val[0].replace('D', 'E')), '%s_orbital' % name: val[1]}

        def str_to_fon(val_in):
            val = [v.split('=') for v in val_in.split('\n')]
            return {v[0].strip(): float(v[1]) for v in val if len(v) == 2}

        def str_to_energies(val_in):
            re_energy = re.compile(rf'([\w\(\) ]+)\s*(?:=|:)\s*({re_float})')
            energies = dict()
            for val in val_in.strip().split('\n'):
                val = re_energy.search(val)
                if val:
                    key = val.group(1).strip()
                    energies[key] = float(val.group(2)) * ureg.hartree if 'energy' in key else float(val.group(2))
            return energies

        def str_to_gw_parameters(val_in):
            val = [v.rsplit(' ', 1) for v in val_in.strip().split('\n')]
            parameters = dict()
            for v in val:
                if v[1] in ['T', 'F']:
                    v[1] = v[1] == 'T'
                parameters[v[0].strip()] = v[1]
            return parameters

        def str_to_qp_states(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            val = np.transpose(np.array([v[1:] for v in val if len(v) == 9], dtype=float))
            keys = [
                'eigenvalue_ks_GroundState', 'eigenvalue_quasiParticle_energy',
                'eigenvalue_ExchangeCorrelation_perturbativeGW', 'eigenvalue_ExactExchange_perturbativeGW',
                'eigenvalue_correlation_perturbativeGW', 'eigenvalue_ks_ExchangeCorrelation',
                'Z_factor', 'ExchangeCorrelation_perturbativeGW_derivation']
            states = dict()
            for n, key in enumerate(keys):
                states[key] = val[n] * ureg.eV if key.startswith('eigenvalue') else val[n]
            return states

        def str_to_options(val_in):
            val = [v.split(':', 1) for v in val_in.strip().split('\n')]
            return {v[0].strip(): v[1].strip() for v in val if len(v) == 2}

        def str_to_convergence(val_in):
            re_value = re.compile(r'([\w ]+)\s*(yes|no)\s*')
            convergence = dict()
            for val in val_in.split('\n'):
                val = re_value.search(val)
                if val:
                    convergence[val.group(1).strip()] = val.group(2) == 'yes'
            return convergence

        def str_to_thermodynamics(val_in):
            keys, vals = val_in.strip().split('\n')
            # we could have taken units from prinout but some quantities do not have units
            keys = [k.strip().lower() for k in keys.split()]
            vals = [float(v) for v in vals.split()]
            units = {
                't': ureg.K, 'p': ureg.MPa, 'chem.pot.': ureg.kJ / MOL, 'energy': ureg.kJ / MOL,
                'entropy': ureg.kJ / MOL, 'cv': ureg.kJ / ureg.K / MOL, 'cp': ureg.kJ / ureg.K / MOL,
                'enthalpy': ureg.kJ / MOL}
            vals = {key: vals[n] * units.get(key, 1.0) for n, key in enumerate(keys)}
            # we calculate entropy energy correction
            if 'entropy' in vals and 't' in vals:
                vals['entropy'] *= vals['t'].magnitude
            return vals

        iteration_quantities = [
            Quantity(
                'damping_scf_iteration',
                rf'current damping\s*:*=*\s*({re_float})',
                str_operation=lambda x: 1.0 / (1.0 + float(x)), convert=False),
            Quantity(
                'energies',
                r'ATION\s*ENERGY.+\s*(\d*\s*.+)',
                str_operation=str_to_scf_energies, convert=False),
            Quantity(
                'energy_XC',
                rf'Exc\s*=\s*({re_float})',
                dtype=float, unit=ureg.hartree),
            Quantity(
                'norm_diis_scf_iteration',
                rf'Norm of current diis error\s*:\s*({re_float})', dtype=float),
            Quantity(
                'norm_fia',
                rf'max\. resid\. (norm for Fia\-block=\s*{re_float})\s*for orbital\s*(\w+)',
                str_operation=str_to_norm, convert=False),
            Quantity(
                'norm_fock',
                rf'max\. resid\. (fock norm\s*=\s*{re_float})\s*for orbital\s*(\w+)',
                str_operation=str_to_norm, convert=False),
            Quantity(
                'delta_eigenvalues',
                rf'Delta Eig\.\s*=\s*({re_float})', dtype=float, unit=ureg.eV),
            Quantity(
                'fon',
                r'FON Calculation\s*\-+\s*([\s\S]+?\-{50})',
                str_operation=str_to_fon, convert=False)]

        scf_quantities = [
            Quantity(
                'iteration',
                rf'(current damping\s*:\s*{re_float}\s*ITERATION[\s\S]+?)\n\s*\n',
                repeats=True, sub_parser=TextParser(quantities=iteration_quantities)),
            # alternative format
            Quantity(
                'iteration',
                r'(ATION\s*ENERGY[\s\S]+?)(?:\n *\n *ITER|End of SCF iterations)',
                repeats=True, sub_parser=TextParser(quantities=iteration_quantities)),
            Quantity(
                'number_of_scf_iterations',
                r'convergence criteria satisfied after\s*(\d+)\s*iterations',
                dtype=int)]

        module_quantities = [
            Quantity(
                'atomic_info',
                r'Atomic coordinate, charge and isotope? information\s*\|\s*\+\-+\+\s*'
                r'([\s\S]+?center of nuclear charge.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'info',
                        r'(atomic\s+coordinates\s+atom(?:\s+shells)?\s+charge(?:\s+pseudo)?\s+isotop\s*)'
                        r'([\s\S]+?)\n\s*\n', str_operation=str_to_atomic_info),
                    Quantity(
                        'center_of_nuclear_mass',
                        rf'center of nuclear mass\s*:\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float),
                    Quantity(
                        'center_of_nuclear_charge',
                        rf'center of nuclear charge\s*:\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float)])),
            Quantity(
                'basis_set_info',
                r'(\w* (?:basis set|BASIS SET) information\s*[\s\S]+?)\n\s*\n *\n',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'auxiliary',
                        r'(\w*) (?:basis|BASIS)',
                        str_operation=lambda x: 'auxiliary' in x.lower(), convert=False),
                    Quantity(
                        'spherical',
                        r'(with spherical basis functions)',
                        str_operation=lambda x: True, convert=False),
                    Quantity(
                        'atom',
                        r'(\w*)\s*(\d+)\s*(\d+)\s*(\d+)\s*([\w\-]+)\s*\[([\w\|]+)\]',
                        repeats=True),
                    Quantity(
                        'total',
                        r'total number of (.+?):\s*(\d+)',
                        repeats=True, str_operation=lambda x: x.rsplit(' ', 1))])),
            Quantity(
                'mo_occupation',
                r'mo occupation\s*:\s*irrep\s*mo\'s\s*occupied\s*([\s\S]+?)\n\s*\n',
                str_operation=lambda x: [v.split() for v in x.strip().split('\n')]),
            Quantity(
                'wavefunction_model',
                r'([w ]+)closed shell calculation for the wavefunction models:\s*(\S+)'),
            Quantity(
                'dft_functional',
                r'density functional\s*\-+(\s*[\s\S]+?)\n\s*\n *\n', sub_parser=TextParser(quantities=[
                    Quantity(
                        'functional',
                        r'(?:functional\s*:\s*(.+)|\n *(\S+)\s*(?:meta-GGA |hybrid )*functional\n)'),
                    Quantity('exchange', r'exchange:\s*(.+)', flatten=False),
                    Quantity('correlation', r'correlation:\s*(.+)', flatten=False)])),
            Quantity(
                'embedding_point_charges',
                r'EMBEDDING IN PERIODIC POINT CHARGES\s*\|\s*.+\s*.+\s*'
                r'([\s\S]+?)\n\s*\n *\n', sub_parser=TextParser(quantities=[
                    Quantity(
                        'pceem_max_multipole',
                        r'Maximum multipole moment used\s*:\s*(\d+)', dtype=int),
                    Quantity(
                        'pceem_multipole_precision',
                        rf'Multipole precision parameter\s*:\s*({re_float})', dtype=float),
                    Quantity(
                        'pceem_min_separation_cells',
                        rf'Minimum separation between cells\s*:\s*({re_float})', dtype=float),
                    Quantity(
                        'lattice_vectors',
                        r'Cell vectors \(au\):\s*(.+)\s*(.+)\s*(.+)', dtype=float, unit=ureg.bohr),
                    Quantity(
                        'redefined',
                        r'Redefined unit cell content \(au\):\s*.+\s*([\s\S]+?)\n\s*\n',
                        str_operation=str_to_embedding_coordinates, convert=False),
                    Quantity(
                        'pc_cluster',
                        r'PC cluster transformed to the center of cell 0 \(au\):\s*.+\s*([\s\S]+?)\n\s*\n',
                        str_operation=str_to_embedding_coordinates, convert=False),
                    Quantity(
                        'qm_cluster',
                        r'QM cluster transformed to the center of cell 0 \(au\):\s*.+\s*([\s\S]+?)\n\s*\n',
                        str_operation=str_to_embedding_coordinates, convert=False)])),
            Quantity(
                'dft_d3',
                r'(DFTD3 V[\d\.]+ Rev \d+[\s\S]+?)\n\s*\n *\n',
                sub_parser=TextParser(quantities=[
                    Quantity('version', r'DFTD3 V([\d\.]+ Rev \d+)', flatten=False),
                    Quantity(
                        'energy_van_der_Waals',
                        rf'Edisp /kcal,au:\s*\S+\s*({re_float})', dtype=float, unit=ureg.hartree)])),
            Quantity(
                'uhf',
                r'UHF mod.+? switched (on) !',
                str_operation=lambda x: x == 'on', convert=False),
            Quantity(
                'smearing',
                r'(\w+ smearing switched on[\s\S]+?)\n\s*\n',
                sub_parser=TextParser(quantities=[
                    Quantity('smearing_kind', r'(\w+) smearing', str_operation=lambda x: x.lower()),
                    Quantity(
                        'smearing_width',
                        rf'Final electron temperature:\s*({re_float})', dtype=float)])),
            Quantity(
                'energy_reference_wavefunction',
                rf'Energy of reference wave function is\s*({re_float})', dtype=float),
            Quantity(
                'energies',
                r'(total energy\s*=[\s\S]+?)\n\s*\n',
                str_operation=str_to_energies, convert=False),
            Quantity(
                'self_consistency',
                r'OPTIMIZATION OF THE GROUND STATE CLUSTER AMPLITUDES\s*\*\s*\*\s*\*\s*\*+'
                r'([\s\S]+?)\*{50}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'(Iter\.\s+CCSD energy[\s\S]+?)(?:CC equations|\n *\n)',
                        str_operation=str_to_iteration),
                    Quantity(
                        'convergence',
                        r'(?:converged in (\d+) iterations|within (\d+) cycles)', dtype=int)])),
            Quantity(
                'energies_MP2',
                r'(RHF\s*energy[\s\S]+?SOS\-MP2 energy.+)', str_operation=str_to_energies),
            Quantity(
                'energies_CCSD',
                r'(RHF\s*energy[\s\S]+?Final CCSD energy.*)', str_operation=str_to_energies),
            Quantity(
                'energies_CCSD(T0)',
                r'(RHF\s*energy[\s\S]+?Final CCSD\(T0\) energy.+)', str_operation=str_to_energies),
            Quantity(
                'energy_gradient',
                r'cartesian gradient of the energy \(hartree/bohr\)\s*\-+\s*([\s\S]+?)'
                r'calculation of the energy gradient finished',
                str_operation=str_to_energy_gradient, convert=False),
            Quantity(
                'eigenvalue_file',
                r'orbitals \$\S+\s*will be written to file (\w+)', repeats=True),
            Quantity(
                'cpu_time',
                rf'total\s*cpu-time\s*:\s*(?:({re_float}) days)*\s*(?:({re_float}) hours)*'
                rf'\s*(?:({re_float}) minutes and)*\s*({re_float}) seconds',
                str_operation=str_to_cpu_time, convert=False)]

        aoforce_quantities = module_quantities + [
            Quantity(
                'hessian',
                r'CARTESIAN FORCE CONSTANT MATRIX \(hartree/bohr\*\*2\)\s*'
                r'.+\s*\-+([\s\S]+?)\n\s*\n *\n',
                str_operation=str_to_hessian, convert=False),
            Quantity(
                'hessian_file',
                r'\*\*\*\s*projected\s+hessian\s+written\s+onto\s+\$hessian\s+\(projected\),\s+file=\<(.+)\>'),
            Quantity(
                'normal_modes_vibrational_frequencies',
                r'(NORMAL MODES and VIBRATIONAL FREQUENCIES \(cm\*\*\(-1\)\)\s*\-+[\s\S]+?)(?:\*{50}|\Z)',
                str_operation=str_to_normal_modes_vibrational_frequencies, convert=False),
            Quantity(
                'normal_modes_file',
                r'\*\*\*\s*normal\s+modes\s+written\s+onto\s+\$vibrational\s+normal\s+modes,\s+file=\<(.+?)\>'),
            Quantity(
                'vibrational_spectrum_file',
                r'\*\*\*\s+vibrational\s+spectroscopic\s+data\s+written\s+onto\$vibrational\s+spectrum\s+file=\<(.+?)\>'),
            Quantity(
                'energy_zero_point',
                rf'\*\s*zero point VIBRATIONAL energy\s*:\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'energy_total',
                rf'\*\s*SCF\-energy\s*:\s*({re_float})', drype=float, unit=ureg.hartree),
            Quantity(
                'energy_current',
                rf'\*\s*SCF \+ E\(vib0\)\s*:\s*({re_float})', dtype=float, unit=ureg.hartree)]

        ccsdf12_quantities = module_quantities + [
            Quantity(
                'energies',
                r'time in triples corr.+\s*\*+([\s\S]+?)\*{50}',
                str_operation=str_to_energies, convert=False)]

        dscf_quantities = module_quantities + [
            Quantity(
                'self_consistency',
                r'(scf convergence criterion : increment of total energy[\s\S]+?iterations)',
                sub_parser=TextParser(quantities=scf_quantities))]

        escf_quantities = module_quantities + [
            Quantity(
                'gw',
                r'(GW version\s*\d+[\s\S]+?)\n\s*\n *\n', sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        r'par[ae]meters:\s*\-+([\s\S]+?)\-{20}',
                        str_operation=str_to_gw_parameters),
                    Quantity(
                        'qp_states',
                        r'orb\s*eps\s*QP\-eps\s*Sigma\s*Sigma_x\s*Sigma_c\s*Vxc\s*Z\s*dS/de'
                        r'([\s\S]+?)\n\s*\n',
                        repeats=True, str_operation=str_to_qp_states, convert=False)]))]

        freeh_quantities = module_quantities + [
            Quantity(
                'thermodynamics',
                r'(T\s*[Pp]\s*.+\n)[\s\S]+?(\d[\d\.\- ]+)',
                repeats=True, str_operation=str_to_thermodynamics, convert=False),
            Quantity(
                'energy_zero_point',
                rf'zero point vibrational energy\s*\-+\s*zpe=\s*({re_float})\s*kJ/mol',
                dtype=float, unit=ureg.kJ / MOL)
        ]

        grad_quantities = module_quantities

        pnoccsd_quantities = module_quantities + [
            Quantity('methods', r' Method  \d+:\s*(\S+)', repeats=True),
            Quantity(
                'energies',
                r'(RHF\s*energy[\s\S]+?Final CCSD\(T\) energy.+)', str_operation=str_to_energies)]

        # TODO verify this
        # the only example I found shows ricc2 quantities are same with ccsdf12
        ricc2_quantities = ccsdf12_quantities + [
            Quantity(
                'energy_total',
                rf'Total\s+Energy\s*:\s*({re_float})', dtype=float, unit=ureg.eV)]

        ridft_quantities = module_quantities + [
            Quantity(
                'self_consistency',
                r'Starting SCF iterations([\s\S]+?)\-{40}',
                sub_parser=TextParser(quantities=scf_quantities))]

        rirpa_quantities = module_quantities + [
            # TODO verify this I cannot find an example file
            Quantity(
                'energies_RPA',
                r'(HXX\+RIRPA\s+total\s+energy\s*=[\s\S]+?HXX\s+total\s+energy\s*=.+)',
                str_operation=str_to_energies)]

        statpt_quantities = module_quantities + [
            Quantity(
                'atomic_info',
                r'ATOM\s*CARTESIAN COORDINATES\s*([\s\S]+?)\*{50}',
                sub_parser=TextParser(quantities=[Quantity(
                    'info',
                    r'([\s\S]+)', str_operation=str_to_atomic_info)])),
            Quantity(
                'options',
                r'Stationary point options\s*\*+\s*\*+\s*([\s\S]+?)\*{40}',
                str_operation=str_to_options),
            Quantity(
                'convergence',
                r'CONVERGENCE INFORMATION\s*Converged\?\s*Value\s*Criterion\s*([\s\S]+?)\*{50}',
                str_operation=str_to_convergence)]

        run_quantities = [
            Quantity(
                'x_turbomole_nodename',
                r'\w+\s*\((.+)\)\s*:\s*TURBO', flatten=False),
            Quantity(
                'program_version',
                r'MOLE V([\d\.]+\s*\([\d ]+\))', flatten=False),
            Quantity(
                'time', r'(\d\d\d\d-\d\d-\d\d\s*\d\d:\d\d:\d\d\.\d\d\d)',
                flatten=False, repeats=True),
            Quantity(
                'module_aoforce',
                r'(a o f o r c e \- program[\s\S]+?(?:force : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=aoforce_quantities)),
            Quantity(
                'module_ccsdf12',
                r'(C C S D F 1 2\s*P R O G R A M[\s\S]+?(?:ccsdf12 : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=ccsdf12_quantities)),
            Quantity(
                'module_dscf',
                r'(d s c f \- program[\s\S]+?(?:dscf : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=dscf_quantities)),
            Quantity(
                'module_escf',
                r'(e s c f[\s\S]+?(?:escf : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=escf_quantities)),
            Quantity(
                'module_freeh',
                r'(f r e e   e n t h a l p y - program[\s\S]+?(?:freeh : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=freeh_quantities)),
            Quantity(
                'module_grad',
                r'(g r a d \- program[\s\S]+?(?:grad : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=grad_quantities)),
            Quantity(
                'module_pnoccsd',
                r'(P N O C C S D  - P R O G R A M[\s\S]+(?:pnoccsd : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=pnoccsd_quantities)),
            Quantity(
                'module_ricc2',
                r'(R I C C 2 \- PROGRAM[\s\S]+?(?:ricc2 : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=ricc2_quantities)),
            Quantity(
                'module_ridft',
                r'(r i d f t\s*DFT program[\s\S]+?(?:ridft : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=ridft_quantities)),
            # TODO implement RIRPA, pnoccsd, freeh
            Quantity(
                'module_rirpa',
                r'(PROGRAM\s+RIRPA[\s\s]+?(?:rirpa : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=rirpa_quantities)),
            Quantity(
                'module_statpt',
                r'(this is S T A T P T[\s\S]+?(?:statpt : all done\s*\*\*\*\*|\Z))',
                sub_parser=TextParser(quantities=statpt_quantities))]

        # necessary to wrap it around a module so we keep the order
        self._quantities = [Quantity(
            'module_run',
            r'(.+?TURBOMOLE (?:V|rev)[\s\S]+?all done\s*\*\*\*\*\s*[\d\.\-: ]+)',
            repeats=True, sub_parser=TextParser(quantities=run_quantities))]


class TurbomoleParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/turbomole', code_name='turbomole',
            code_homepage='https://www.turbomole.org/',
            mainfile_contents_re=(r'Copyright \(C\) [0-9]+ TURBOMOLE GmbH, Karlsruhe'))
        self._metainfo_env = m_env
        self.out_parser = OutParser()
        self.matrix_parser = AuxiliaryOutParser(r' *(\d+\s+\d+\s+[\d\.\- ]+)\n', 2)
        self.vibrational_parser = AuxiliaryOutParser(
            r' *(\d+\s+\w*\s+[\d\.\-]+\s+[\d\.\-]+\s+[YES\-]+\s+[YES\-]+)', 0)
        self.eigenvalues_parser = EigenvaluesParser()
        self.control_parser = ControlParser()

        self.xc_functional_map = {
            'S-VWN': ['LDA_X', 'LDA_C_VWN_3'],
            'PWLDA': ['LDA_X', 'LDA_C_PW'],
            'B-VWN': ['GGA_X_B88', 'LDA_C_VWN'],
            'B-LYP': ['GGA_X_B88', 'GGA_C_LYP'],
            'B-P': ['GGA_X_B88', 'GGA_C_P86'],
            'B-P86': ['GGA_X_B88', 'GGA_C_P86'],
            'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'TPSS': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            'M06': ['MGGA_X_M06', 'MGGA_C_M06'],
            'BH-LYP': ['HYB_GGA_XC_BHANDHLYP'],
            'B3-LYP': ['HYB_GGA_XC_B3LYP'],
            'PBE0': ['HYB_GGA_XC_PBEH'],
            'TPSSh': ['HYB_MGGA_XC_TPSSH'],
            'M06-2X': ['MGGA_X_M06_2X', 'MGGA_C_M06_2X'],
            'B2-PLYP': ['HYB_GGA_XC_B2PLYP']}

        self.models_map = {
            'MP2': 'MP2', 'CCS': 'CCS', 'CIS': 'CIS', 'CIS(D)': 'CISD', 'CIS(Dinf)': 'CISD',
            'ADC(2)': 'MP2',  # TODO: check paper to verify this mapping
            'CC2': 'CCSD', 'CCSD': 'CCSD', 'CCSD(T)': 'CCSD(T)'}

        self.metainfo_map = {
            'total energy': 'energy_total', 'kinetic energy': 'electronic_kinetic_energy',
            'potential energy': 'x_turbomole_potential_energy_final',
            'virial theorem': 'x_turbomole_virial_theorem', 'wavefunction norm': 'x_turbomole_wave_func_norm',
            'Final CCSD(T) energy': 'energy_total', 'CCSD correlation energy': 'energy_current',
            'Final CCSD(T0) energy': 'energy_total',
            'Final CCSD energy': 'energy_total', 'correlation energy': 'energy_current',
            'Final MP2 energy': 'energy_total', 'MP2 correlation energy (doubles)': 'energy_current',
            r'HXX\+RIRPA total energy': 'energy_total', 'RIRPA correlation energy': 'energy_current',
            'HXX total energy': 'energy_total', 'rpa response function': 'x_turbomole_gw_use_rpa_response',
            'eta (Hartree)': 'x_turbomole_gw_eta_factor',
            'Maximum allowed trust radius': 'x_turbomole_geometry_optimization_trustregion_max',
            'Minimum allowed trust radius': 'x_turbomole_geometry_optimization_trustregion_min',
            'Initial trust radius': 'x_turbomole_geometry_optimization_trustregion_initial',
            'Hessian update method': 'geometry_optimization_method',
            'Threshold for energy change': 'geometry_optimization_energy_change',
            'Threshold for max displacement element': 'geometry_optimization_geometry_change',
            'Threshold for max gradient element': 'geometry_optimization_threshold_force',
            'Threshold for RMS of displacement': 'x_turbomole_geometry_optimization_geometry_change_rms',
            'Threshold for RMS of gradient': 'x_turbomole_geometry_optimization_threshold_force_rms',
            't': 'temperature', 'p': 'pressure', 'entropy': 'energy_correction_entropy',
            'energy': 'energy_total', 'cv': 'heat_capacity_C_v', 'cp': 'heat_capacity_C_p'}

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self.module = None

    def get_number_of_atoms(self):
        return len(self.module.get('atomic_info', {}).get('info', {}).get('atom', []))

    def get_number_of_spin_channels(self):
        nspin = 1
        if self.module.get('uhf'):
            nspin = 2
        gw_parameters = self.module.get('gw', {}).get('parameters', None)
        if gw_parameters is not None:
            nspin = gw_parameters.get('number of spin channels', 1)
        # TODO add other conditions
        return nspin

    def get_electronic_structure_method(self):
        if self.module.get('wavefunction_model') is not None:
            return self.models_map.get(self.module.get('wavefunction_model'), None)
        if self.module.get('methods') is not None:
            return self.models_map.get(self.module.get('methods')[-1], None)
        if self.module.get('dft_functional') is not None:
            return 'DFT'
        if self.module.get('gw') is not None:
            return 'G0W0'

    def parse_system(self):
        sec_run = self.archive.section_run[0]
        sec_system = sec_run.m_create(System)

        info = self.module.get('atomic_info', {}).get('info', {})
        if info.get('coordinates') is not None:
            sec_system.atom_positions = info.get('coordinates', [])
        if info.get('atom') is not None:
            sec_system.atom_labels = info.get('atom', [])
        sec_system.configuration_periodic_dimensions = [False, False, False]
        # TODO is it always non-periodic?

        point_charges = self.module.get('embedding_point_charges', {})
        cluster_description = {
            'redefined': 'periodic point charges for embedding',
            'pc_cluster': 'removed point charge cluster', 'qm_cluster': 'shifted embedded QM cluster'}
        for name, description in cluster_description.items():
            cluster = point_charges.get(name, None)
            if cluster is None:
                continue
            sec_system_cluster = sec_run.m_create(System)
            sec_system_cluster.atom_positions = cluster.get('coordinates')
            sec_system_cluster.atom_labels = cluster.get('labels')
            if name == 'redefined':
                sec_system_cluster.x_turbomole_pceem_charges = cluster.get('charges')
                sec_system_cluster.lattice_vectors = point_charges.get('lattice_vectors')
            sec_system_cluster.configuration_periodic_dimensions = [True, True, True]
            # TODO is the reference correct or the other way around?
            sec_system_refs = sec_system.m_create(SystemToSystemRefs)
            sec_system_refs.system_to_system_ref = sec_system_cluster
            sec_system_refs.system_to_system_kind = description

        for key, val in point_charges.items():
            if key.startswith('pceem_') and val is not None:
                setattr(sec_system, 'x_turbomole_%s' % key, val)

        return sec_system

    def parse_scc(self):
        sec_run = self.archive.section_run[0]
        sec_scc = sec_run.m_create(SingleConfigurationCalculation)

        n_atoms = self.get_number_of_atoms()

        # energies
        for key, val in self.module.get('energies', {}).items():
            key = self.metainfo_map.get(key, None)
            if key is None:
                continue
            setattr(sec_scc, key, val)

        for key in ['zero_point', 'current', 'total']:
            key = 'energy_%s' % key
            val = self.module.get(key)
            if val is not None:
                setattr(sec_scc, key, val.to('joule').magnitude)

        # module-specific energies (will create additional scc, method)
        for name, energies in self.module.items():
            if not name.startswith('energies_') or energies is None:
                continue
            sec_scc_module = sec_run.m_create(SingleConfigurationCalculation)
            for key, val in energies.items():
                key = self.metainfo_map.get(key, None)
                if key is None:
                    continue
                setattr(sec_scc_module, key, val)
            sec_method_module = sec_run.m_create(Method)
            sec_method_module.electronic_structure_method = name.lstrip('energies_')
            sec_scc_module.single_configuration_to_calculation_method_ref = sec_method_module
            # add reference to main sec_scc
            sec_scc_refs = sec_scc_module.m_create(CalculationToCalculationRefs)
            sec_scc_refs.calculation_to_calculation_ref = sec_scc
            sec_scc_refs.calculation_to_calculation_kind = 'starting_point'

        # forces
        energy_gradient = self.module.get('energy_gradient')
        if energy_gradient is not None:
            sec_scc.atom_forces_raw = energy_gradient

        # thermodynamics
        thermodynamics = self.module.get('thermodynamics')
        if thermodynamics is not None:
            current_t, current_p = 0., 0.
            sec_scc_thermo = None
            for thermo in thermodynamics:
                if current_t != thermo.get('t', 0.) or current_p != thermo.get('p', 0.):
                    sec_scc_thermo = sec_run.m_create(SingleConfigurationCalculation)
                    sec_scc_refs = sec_scc_thermo.m_create(CalculationToCalculationRefs)
                    sec_scc_refs.calculation_to_calculation_ref = sec_scc
                    sec_scc_refs.calculation_to_calculation_kind = 'starting_point'
                current_t, current_p = thermo.get('t', 0.), thermo.get('p', 0.)
                if sec_scc_thermo is not None:
                    for key, val in thermo.items():
                        key = self.metainfo_map.get(key, key)
                        setattr(sec_scc_thermo, key, val)

        # hessian
        if self.module.get('hessian') is not None:
            sec_scc.hessian_matrix = self.module.get('hessian').to('J/m**2').magnitude
        elif self.module.get('hessian_file') is not None:
            # read if from file
            self.matrix_parser.mainfile = os.path.join(self.maindir, self.module.get(
                'hessian_file'))
            try:
                hessian = np.reshape(np.array(
                    self.matrix_parser.matrix, dtype=float), (n_atoms, 3, n_atoms, 3))
                hessian = np.transpose(hessian, axes=(
                    0, 2, 1, 3)) * ureg.hartree / ureg.bohr**2
                sec_scc.hessian_matrix = hessian.to('J/m**2').magnitude
            except Exception:
                self.logger.warn('Cannot read hessian file.')

        # vibrational frequencies
        vibration = self.module.get('normal_modes_vibrational_frequencies')
        if vibration is not None:
            for key, val in vibration.items():
                setattr(sec_scc, 'x_turbomole_vibrations_%s' % key, val)
        else:
            if self.module.get('normal_modes_file') is not None:
                # read from file
                self.matrix_parser.mainfile = os.path.join(self.maindir, self.module.get(
                    'normal_modes_file'))
                try:
                    normal_modes = np.array(self.matrix_parser.matrix, dtype=float)
                    sec_scc.x_turbomole_vibrations_normal_modes = np.transpose(np.reshape(
                        normal_modes, (n_atoms, 3, len(normal_modes[0]))), axes=(2, 0, 1))
                except Exception:
                    self.logger.warn('Cannot read normal modes file.')

            if self.module.get('vibrational_spectrum_file') is not None:
                self.vibrational_parser.mainfile = os.path.join(self.maindir, self.module.get(
                    'vibrational_spectrum_file', ''))
                try:
                    spectrum = self.vibrational_parser.matrix
                    sec_scc.x_turbomole_vibrations_mode_energies = np.array(
                        [s[-4] for s in spectrum], dtype=float)
                    sec_scc.x_turbomole_vibrations_intensities = np.array(
                        [s[-3] for s in spectrum], dtype=float)
                    sec_scc.x_turbomole_vibrations_infrared_activity = [s[-2] == 'YES' for s in spectrum]
                    sec_scc.x_turbomole_vibrations_raman_activity = [s[-1] == 'YES' for s in spectrum]

                except Exception:
                    self.logger.warn('Cannot read vibrational spectrum file.')

        # eigenvalues
        eigenvalue_files = self.module.get('eigenvalue_file')
        if eigenvalue_files is not None:
            eigenvalues = []
            occupation = []
            irrep = []
            try:
                for eigenvalue_file in eigenvalue_files:
                    self.eigenvalues_parser.mainfile = os.path.join(self.maindir, eigenvalue_file)
                    eigenvalues.append(np.hstack(self.eigenvalues_parser.get('eigenvalues', [])))
                    occupation.append(np.hstack(self.eigenvalues_parser.get('occupation', [])))
                    irrep.append(np.hstack(self.eigenvalues_parser.get('irrep', [])))
                    sec_eigenvalues = sec_scc.m_create(Eigenvalues)
                    sec_eigenvalues.eigenvalues_values = np.reshape(np.array(
                        eigenvalues, dtype=float), (len(eigenvalue_files), 1, len(eigenvalues[0]))) * ureg.hartree
                    sec_eigenvalues.eigenvalues_occupation = np.reshape(np.array(
                        occupation, dtype=float), (len(eigenvalue_files), 1, len(occupation[0])))
                    sec_eigenvalues.x_turbomole_eigenvalues_irreducible_representation = np.reshape(np.array(
                        irrep, dtype=np.dtype('U')), (len(eigenvalue_files), 1, len(irrep[0])))
                    sec_eigenvalues.eigenvalues_kpoints = [np.zeros(3)]
            except Exception:
                self.logger.warn('Cannot read eigenvalues.')

        # basis set
        for basis_set in self.module.get('basis_set_info', []):
            basis_set_mapping = dict()
            for atom in basis_set.get('atom', []):
                symbol = atom[0].title()
                sec_atom_basis = sec_run.m_create(BasisSetAtomCentered)
                sec_atom_basis.basis_set_atom_centered_short_name = atom[4]
                sec_atom_basis.basis_set_atom_number = atomic_numbers.get(symbol, 0)
                if basis_set.spherical:
                    sec_atom_basis.number_of_basis_functions_in_basis_set_atom_centered = atom[3]
                basis_set_mapping[symbol] = sec_atom_basis

            # all atoms with duplicate in basis_set_info
            if len(basis_set.get('atom', [])) == self.get_number_of_atoms():
                sec_basis_set = sec_scc.m_create(BasisSet)
                sec_basis_set.mapping_section_basis_set_atom_centered = [
                    basis_set_mapping[symbol] for symbol in self.module.get(
                        'atomic_info', {}).get('info', {}).get('atom', [])]
                sec_basis_set.basis_set_kind = 'density' if basis_set.auxiliary else 'wavefunction'

        # self consistency
        self_consistency = self.module.get('self_consistency', {})
        # total energies for calculation of energy change
        energies = [0.] + [iteration.get('energies', {}).get(
            'energy_total_scf_iteration') for iteration in self_consistency.get('iteration', [])]
        for n, iteration in enumerate(self_consistency.get('iteration', [])):
            sec_iteration = sec_scc.m_create(ScfIteration)

            for key, val in iteration.get('energies', {}).items():
                setattr(sec_iteration, key, val)
            # energy change
            sec_iteration.energy_change_scf_iteration = energies[n]
            # scf quantities
            for key in ['energy_total', 'energy_XC', 'time']:
                val = iteration.get(key, None)
                if val is None:
                    continue
                setattr(sec_iteration, '%s_scf_iteration' % key, val)
            # miscellaneous scf quantities
            scf_keys = [
                'damping_scf_iteration', 'norm_diis_scf_iteration', 'delta_eigenvalues',
                'norm_fia', 'norm_fock']
            for key in scf_keys:
                val = iteration.get(key, None)
                if val is None:
                    continue
                if key == 'norm_fia' or key == 'norm_fock':
                    setattr(sec_iteration, 'x_turbomole_%s_orbital_scf_iteration' % key, val.get('%s_orbital' % key))
                    val = val.get(key)
                    key = '%s_scf_iteration' % key
                setattr(sec_iteration, 'x_turbomole_%s' % key, val)
        n_scf = self_consistency.get('number_of_scf_iterations', None)
        if n_scf is not None:
            sec_scc.number_of_scf_iterations = n_scf

        # gw
        if self.module.get('gw') is not None:
            sec_eigenvalues = sec_scc.m_create(Eigenvalues)
            for qp_states in self.module.gw.get('qp_states', []):
                sec_gw_eigenvalues = sec_eigenvalues.m_create(x_turbomole_section_eigenvalues_GW)
                for key, val in qp_states.items():
                    setattr(sec_gw_eigenvalues, 'x_turbomole_%s' % key, val)

        # vdW
        energy_vdW = self.module.get('dft_d3', {}).get('energy_van_der_Waals')
        if energy_vdW is not None:
            sec_vdW = sec_scc.m_create(EnergyVanDerWaals)
            sec_vdW.energy_van_der_Waals_kind = 'DFTD3'
            sec_vdW.energy_van_der_Waals = energy_vdW

        return sec_scc

    def parse_method(self):
        sec_method = self.archive.section_run[0].m_create(Method)

        method = self.get_electronic_structure_method()
        if method is not None:
            sec_method.electronic_structure_method = method

        sec_method.number_of_spin_channels = self.get_number_of_spin_channels()

        for key, val in self.module.get('smearing', {}).items():
            setattr(sec_method, key, val)

        atomic_info = self.module.get('atomic_info')
        if atomic_info is not None:
            info = atomic_info.get('info', {})
            atom_kind = list(set(info.get('atom', [])))
            for atom in atom_kind:
                sec_atom_kind = sec_method.m_create(MethodAtomKind)
                sec_atom_kind.method_atom_kind_atom_number = atomic_numbers.get(atom, 0)
                sec_atom_kind.method_atom_kind_label = atom

            # only atom kinds
            for basis_set in self.module.get('basis_set_info', []):
                atom = basis_set.get('atom', [])
                if len(atom) != len(atom_kind):
                    continue
                basis_set_mapping = {a[0].title(): n for n, a in enumerate(atom)}
                sec_basis_set = sec_method.m_create(MethodBasisSet)
                sec_basis_set.number_of_basis_sets_atom_centered = len(basis_set.get('atom', []))
                sec_basis_set.method_basis_set_kind = 'density' if basis_set.auxiliary else 'wavefunction'
                sec_basis_set.mapping_section_method_basis_set_atom_centered = [[
                    n, basis_set_mapping[symbol]] for n, symbol in enumerate(atom_kind)]

        # XC Functionals
        dft_functional = self.module.get('dft_functional')
        if dft_functional is not None:
            sec_method.calculation_method_kind = 'absolute'
            for functional in self.xc_functional_map.get(dft_functional.get('functional'), ['HF_X']):
                sec_xc_functional = sec_method.m_create(XCFunctionals)
                sec_xc_functional.XC_functional_name = functional

            if dft_functional.exchange is not None:
                sec_method.x_turbomole_functional_type_exchange = dft_functional.exchange

            if dft_functional.correlation is not None:
                sec_method.x_turbomole_functional_type_correlation = dft_functional.correlation

        # vdW method
        dftd3_version = self.module.get('dft_d3', {}).get('version', None)
        if dftd3_version is not None:
            sec_method.van_der_Waals_method = 'DFT-D3'
            sec_method.x_turbomole_dft_d3_version = dftd3_version

        if self.module.get('gw') is not None:
            sec_method.calculation_method_kind = 'perturbative'
            parameters = self.module.gw.get('parameters', {})
            for key, val in parameters.items():
                val = val * ureg.hartree if 'Hartree' in key else val
                key = self.metainfo_map.get(key, None)
                if key is None:
                    continue
                setattr(sec_method, key, val)
            gw_type = parameters.get('type of gw 0:G0W0 1: GW0', 0)
            sec_method.x_turbomole_gw_approximation = ['G0W0', 'GW0'][gw_type]

        # control
        self.control_parser.mainfile = os.path.join(self.maindir, 'control')
        for key, val in self.control_parser.items():
            if val is None:
                continue
            if key == 'drvopt':
                for drvopt_key, drvopt_val in val.items():
                    setattr(sec_method, 'x_turbomole_controlIn_%s_status' % drvopt_key, drvopt_val)
            else:
                setattr(sec_method, 'x_turbomole_controlIn_%s' % key, val)

        return sec_method

    def parse_sampling_method(self):
        options = self.module.get('options')
        if options is None:
            return

        sec_sampling = self.archive.section_run[0].m_create(SamplingMethod)
        for key, val in options.items():
            if 'radius' in key or 'displacement' in key:
                val = val * ureg.bohr
            elif 'energy' in key:
                val = val * ureg.hartree
            elif 'gradient' in key:
                val = val * ureg.hartree / ureg.bohr
            key = self.metainfo_map.get(key, None)
            if key is None:
                continue
            setattr(sec_sampling, key, val)

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program_name = 'turbomole'
        sec_run.program_basis_set_type = 'gaussians'

        for header in ['program_version', 'x_turbomole_nodename']:
            value = [module_run.get(header) for module_run in self.out_parser.get('module_run', [])]
            if len(set(value)) > 1:
                self.logger.warn('Multiple values found for header.')
            if value:
                setattr(sec_run, header, value[0])

        time = [module_run.get('time') for module_run in self.out_parser.get('module_run', [])]
        time = [t for t in time if len(t) == 2]
        start = datetime.strptime(time[0][0], '%Y-%m-%d %H:%M:%S.%f') - datetime.utcfromtimestamp(0)
        sec_run.time_run_date_start = start.total_seconds()
        end = datetime.strptime(time[-1][1], '%Y-%m-%d %H:%M:%S.%f') - datetime.utcfromtimestamp(0)
        sec_run.time_run_date_end = end.total_seconds()

        for module_run in self.out_parser.get('module_run', []):
            for name, module in module_run.items():
                if name.startswith('module_') and module is not None:
                    self.module = module

                    sec_method = self.parse_method()
                    sec_system = self.parse_system()
                    sec_scc = self.parse_scc()
                    self.parse_sampling_method()
                    sec_scc.single_configuration_to_calculation_method_ref = sec_method
                    sec_scc.single_configuration_calculation_to_system_ref = sec_system
